from builtin.engtool import *

# > sort 6
ENGINE_NAME = "ATEngine"


# 定义一个元类，用于自动注册类
class _UpdatableMeta(type):
    # 当一个类被定义时，它将通过 scheduler.add(cls) 被注册
    def __new__(meta, name, bases, attrs):
        # 创建类
        new_class = type.__new__(meta, name, bases, attrs)
        # LOG_OUTPUT.append(f"new_class: {name}, {new_class}")
        if name != "Updatable" and name != "CreepLogic":
            # 将新创建的类添加到全局的 Scheduler 中
            # print("new_class: ", new_class)
            Scheduler().login(new_class)
        return new_class


class _Updatable:
    __kind__ = "updatable"
    """
    这是一个用于构造可更新对象的基类 | This is a base class for constructing updatable objects
    """

    #   Used to record subclasses of all Updatable(only considering those defined in Src)
    def __init__(self, _loading: bool = True, _enable: bool = None):
        self._alive = True  # 是否存活. 一旦被置位为False，将不再被更新 | Whether alive. Once set to False, it will no longer be updated
        self._inited = False  # 是否已经初始化 | Whether initialized
        self._enabled = True if _enable is None else _enable  # 被置位时，才允许更新事件 | When set, update events are allowed
        if _loading: self.loading(know)
        self.sch = Scheduler()
        # self.sch.add(self)

    def loading(self, k: GlobalKnowledge) -> any:
        """
        当__init__函数被调用后，执行此函数 | called after the __init__ function is called

        Args:
            k: GlobalKnowledge

        Returns:
            not care
        """
        pass

    def birth(self, k: GlobalKnowledge) -> any:
        """
        * 首次执行step函数前，执行此函数 | called before executing the step function for the first time
        Args:
            k: GlobalKnowledge

        Returns:
            None
        """
        pass

    def step(self, k: GlobalKnowledge) -> any:
        """
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function
        Args:
            k: GlobalKnowledge

        Returns:
            None
        """
        pass

    def removed(self, k: GlobalKnowledge) -> any:
        """
        * 由外部停止此对象时调用此 | called when this object is stopped externally
        Args:
            k: GlobalKnowledge

        Returns:
            None
        """
        pass

    # ==================================================================================
    def disable(self) -> None:
        """
        * 由用户调用，停用此Updatable | called by user to disable this Updatable

        Args:

        Returns:
            None
        """
        self._enabled = False

    def enable(self) -> None:
        """
        * 由用户调用，重新启用此Updatable | called by user to re-enable this Updatable

        Args:

        Returns:
            None
        """
        self._enabled = True

    def remove(self) -> None:
        """
        * 由用户调用，删除此Updatable | called by user to remove this Updatable

        Args:

        Returns:
            None
        """
        if not self._alive: return
        self._alive = False
        self._enabled = False
        self.removed(know)

    @property
    def alive(self) -> bool:
        """
        * 检查此对象是否存活 | called by user to check if this object is alive
        """
        return self._alive

    @property
    def enabled(self) -> bool:
        """
        * 检查此对象是否在运行 | called by user to check if this object is running
        """
        return self._enabled

    # ==================================================================================

    def handle(self) -> bool:
        """
        * 由用户调用，执行此函数 | called by user to update

        Returns:
            bool # 是否成功执行更新逻辑 | whether executed init/step logic successfully
        """

        # 检查是否暂停 | Check if pause
        if not self._enabled:
            return False

        # 执行更新逻辑 | exec Update=
        if not self._inited:
            try:
                self.birth(know)
                self._inited = True
            except Exception as e:
                std.error(f"{self.__kind__}.init", e)

        try:
            self.step(know)
        except Exception as e:
            std.error(f"{self.__kind__}.step", e)

        return True


class _ListenSystem:
    __listens__ = {}  # ename: {.name: func}

    def listenFor(self, ename: str, func: callable):
        if ename not in _ListenSystem.__listens__:
            _ListenSystem.__listens__[ename] = {}
        _ListenSystem.__listens__[ename][self.name] = func

    def cancelListen(self, ename: str):
        if ename in _ListenSystem.__listens__:
            clstype = _ListenSystem.__listens__[ename]
            if self.name in clstype:
                del clstype[self.name]
            if len(clstype) == 0:
                del _ListenSystem.__listens__[ename]

    def pushEvent(self, ename: str, data: dict):
        data["src"] = self.name
        if ename in _ListenSystem.__listens__:
            for func in _ListenSystem.__listens__[ename].values():
                try:
                    func(data)
                except Exception as e:
                    std.warn("ListenSystem", f"Error in {self.name}.listen-for {ename}:", str(e))


class _CreepLogicSettings:
    # ---------------------- 爬虫本体 | Creep --------------------------- #
    NAME: str = "creep"  # Creep名称 | Creep name

    # ---------------------- 流程与结构 | Flow & Struct --------------------------- #
    LINK: list[str] = None  # CreepLogic链接 | CreepLogic link - 在此处标记的link只会启用一次，即使你反复初始化
    ONCE: bool = False  # 如果killed函数被调用，置为True，CreepLogic将被删除 | If the killed function is called, set it to True, the CreepLogic will be deleted

    # ---------------------- 孵化选项 | Spawning --------------------------- #
    RECIPE: list[str] = [MOVE]  # Creep配方 | Creep recipe
    OPTIMISE: bool = True  # 是否自动优化配方 | Whether to automatically optimize the recipe
    EXTENSION: bool = True  # 生产Creep时是否使用extension | Whether to use extension when producing Creep


class CreepLogic(_CreepLogicSettings, _ListenSystem, _Updatable, metaclass=_UpdatableMeta):
    __kind__ = "creep-logic"
    __names__ = {}  # 用于记录CreepLogic的名称 | Used to record the name of CreepLogic
    __types__ = {}  # 用于记录CreepLogic的类型 | Used to record the type of CreepLogic
    __id__ = {}

    def __init__(self, spawn: StructureSpawn = None):
        self.creep = None
        self.spawn = spawn
        self._nexts = []
        self._link_wait = 1  # link会被推迟到下一个tick执行
        self._next_refs = []
        self._name, self.recipe, self.optimise = self.NAME, self.RECIPE, self.OPTIMISE
        self._cost, self._extension, self._once = get.cost(self.recipe), self.EXTENSION, self.ONCE
        self._id = self._get_id()
        if self.__class__.LINK:
            self.link(*self.__class__.LINK)
            self.__class__.LINK = None
        self.name = self._name + str(self._id)
        if not self.recipe:
            std.error("CreepLogic", "CreepLogic must has recipe.")

        name_len = len(self._name)
        if name_len <= 0:
            std.error("CreepLogic", "CreepLogic must has not empty name.")
        if self._name[name_len - 1].isdigit():
            std.error("CreepLogic", "CreepLogic name can not end with a number: " + self._name)
        if self._name in CreepLogic.__types__ and self.__class__.__name__ is not CreepLogic.__types__[self._name]:
            std.error("CreepLogic", "CreepLogic must has unique type: " + self.__class__.__name__)
        else:
            CreepLogic.__types__[self._name] = self.__class__.__name__
        if self._name not in CreepLogic.__names__:
            CreepLogic.__names__[self._name] = {}
        CreepLogic.__names__[self._name][self._id] = self
        _Updatable.__init__(self, self._id is 0, True)
        self._sp_planner = SpawnPlanner()

    def _get_id(self):
        if self._name not in CreepLogic.__id__:
            CreepLogic.__id__[self._name] = -1
        CreepLogic.__id__[self._name] += 1
        return CreepLogic.__id__[self._name]

    def _seek_id(self):
        if self._name not in CreepLogic.__id__:
            return -1
        return CreepLogic.__id__[self._name]

    def link(self, *next_logics: str):
        """
        * 设置下一个CreepLogic | Set the next CreepLogic
        这个函数会在全局类型表中找到对应NAME的CreepLogic，并将其实例化 |
            This function will find the CreepLogic corresponding to NAME in the global type table and instantiate it

        Args:
            next_logics: str  # CreepLogic.NAME

        Returns:
            None
        """
        self._link_wait = 1
        self._nexts.extend(next_logics)

    def _enable_next(self):
        for _type_name in self._nexts:
            _inst = self.sch.add(_type_name)
            self._next_refs.append(_inst)
            _inst.enable(self.spawn)
        self._nexts = []

    @classmethod
    def ref(cls, name: str) -> list | "CreepLogic":
        """
        * 通过名称获取CreepLogic实例 | Get CreepLogic instance by name
        如果你的name不以数字结尾，将返回一个列表，否则返回一个CreepLogic实例 |
            If your name does not end with a number, a list will be returned, otherwise a CreepLogic instance will be returned
        Args:
            name: str  # CreepLogic名称 | CreepLogic name

        Returns:
            list | CreepLogic  # CreepLogic实例或列表 | CreepLogic instance or list

        """
        if st.na(name): return None
        name_len = len(name)
        if name_len <= 0:
            std.error("CreepLogic.ref", "CreepLogic name must not be empty.")

        if not name[name_len - 1].isdigit():
            insts = cls.__names__.py_get(name, {})
            insts = list(insts.values())
            return insts
        else:
            if name_len <= 1:
                std.error("CreepLogic.ref", "CreepLogic name can not be only number: " + name)
            # reversed search a not digit
            for i in range(name_len - 2, -1, -1):
                if not name[i].isdigit():
                    name, num = name[:i + 1], int(name[i + 1:])
                    insts = cls.__names__.py_get(name, {})
                    insts = list(insts.values())
                    if num < len(insts):
                        return insts[num]
                    else:
                        return None
            return None

    def birth(self, c: Creep, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        当自身对应RECIPE的Creep被创建完成时调用此 | called when the Creep corresponding to RECIPE is created

        Args:
            c: Creep
            k: GlobalKnowledge
            refs: CreepLogic

        Returns:
            not care
        """
        pass

    def step(self, c: Creep, k: GlobalKnowledge, *refs: "CreepLogic") -> any:
        """
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function

        Args:
            c: Creep
            k: GlobalKnowledge
            refs: CreepLogic

        Returns:
            not care
        """
        pass

    def killed(self, k: GlobalKnowledge) -> any:
        """
        * 当绑定的Creep被击杀时调用此 | called when the bound Creep is killed

        Args:
            k: GlobalKnowledge

        Returns:
            not care
        """
        pass

    def enable(self, spawn: StructureSpawn = None):
        if not spawn and not self.spawn:
            std.error("CreepLogic", "CreepLogic __init__ without spawn must enable with a spawn .")
        elif spawn:
            self.spawn = spawn
        # print("enable target: ", self.name)
        super().enable()

    @property
    def ready(self) -> bool:
        """
        * 检查该logic是否已经完全就绪 | Check if the logic is ready

        Returns:
            bool
        """
        return self._alive and self._enabled and self._inited and st.creep(self.creep)

    @classmethod
    def productQueue(cls) -> list["CreepLogic"]:
        """
        这个函数可以返回当前等待生产creep的CreepLogic的实例列表 | This function can return a list of instances of CreepLogic waiting to produce creep

        Returns:
            list["CreepLogic"]
        """
        res = []
        sch = Scheduler()
        for logic in sch.updates:
            if st.creep(logic.creep):
                continue
            res.append(logic)
        return res

    def jumpQueue(self):
        """
        * 由用户调用，将此CreepLogic插入到队列的最前面 | called by user to insert this CreepLogic into the front of the queue
        """
        queues = self.productQueue()
        if len(queues) > 0:
            # 获取spawn是否忙
            if not st.spawn(self.spawn):
                if LANGUAGE == "cn":
                    std.warn("CreepLogic.jumpQueue", "Not bind a spawn.")
                else:
                    std.warn("CreepLogic.jumpQueue", "Not bind a spawn.")
                return
            if self.spawn.spawning:
                if self.spawn.spawning.creep.name == self.name:
                    return  # 已经在生产了 | Already in production
                elif len(queues) > 1:
                    target = queues[1]  # 有多个CreepLogic在等待，插入到第二个 | There are multiple CreepLogic waiting, insert into the second
                else:
                    return  # 只有一个CreepLogic在等待，不插入 | Only one CreepLogic is waiting, do not insert
            else:
                target = queues[0]
        else:
            return

        idx = self.sch.updates.indexOf(target)
        if idx >= 0:
            # > insert self.sch.updates.remove(self)
            # > insert self.sch.updates.splice(idx, 0, self)
            pass

    def handle(self) -> bool:
        """
        * 由用户调用，执行此函数 | called by user to update

        Returns:
            bool # 是否成功执行更新逻辑 | whether executed init/step logic successfully
        """

        # 检查是否暂停 | Check if pause
        if not self._enabled:
            return False

        # 检查Creep状态
        if not st.creep(self.creep, False):
            if not self.recipe:  # 表明这是一个一次性的CreepLogic | This is a one-time CreepLogic
                self.remove()
                return False
            elif self.spawn:  # 尝试创建新的Creep | Try to create a new Creep
                if not self._extension:
                    if get.energy(self.spawn) < self._cost:
                        return False

                self.creep = self._sp_planner.require(self)
                if st.jnumber(self.creep) and self.creep > 0:
                    self._extension = True  # 重置为True，表示不再限制extension | Reset to True, indicating no longer limiting extension
                if not st.creep(self.creep):
                    return False
                self._inited = False
            else:
                std.error("CreepLogic", "CreepLogic must __init__ with a creep or has recipe and a spawn.")
        elif not self.creep.exists:
            self.killed(know, *self._next_refs)
            self.creep = None
            if self._once: self.remove()
            return False

        # 执行更新逻辑 | exec Update=
        if not self._inited:
            try:
                self.birth(self.creep, know, *self._next_refs)
                self._inited = True
            except Exception as e:
                std.error(f"{self.__kind__}.init", e)

        # next
        try:
            if self._nexts:
                self._enable_next()
        except Exception as e:
            std.error(f"{self.__kind__}.enable_next", e)

        # step
        try:
            self.step(self.creep, know, *self._next_refs)
        except Exception as e:
            std.error(f"{self.__kind__}.step", e)

        return True

    def __str__(self):
        if st.creep(self.creep):
            creep_string = f"[Creep({self.creep.name})]"
        else:
            creep_string = "[Creep(None)]"
        return f"CreepLogic({self.name}) {creep_string};"

class SitePlan:
    def __init__(self, ref: Point, dx:int, dy:int, building_type: type):
        """
        初始化一个建筑规划点

        Args:
            ref (Point): 参考点
            dx (int): 相对于参考点的x轴偏移量
            dy (int): 相对于参考点的y轴偏移量
            building_type (type): 建筑类型
        """
        self.ref = ref
        self.offset = Point(dx, dy)
        self.building_type = building_type
        self.site = None

    def prepare(self) -> int:
        """
        准备建筑点

        计算实际建筑点位置，并尝试创建建筑点

        Returns:
            int: 返回创建建筑点的结果标识
        """
        self.point = Point(self.ref.x + self.offset.x, self.ref.y + self.offset.y)
        self.site = put.site(self.point.x, self.point.y, self.building_type)
        return self.site

    def __str__(self):
        return f"SitePlan({self.point}, {self.building_type})"


class SitePlaner:
    def __init__(self, plans: list[SitePlan]):
        """
        初始化一个建筑规划器

        Args:
            plans (list[SitePlan]): 建筑规划点列表
        """
        self.plans = plans
        self.site:ConstructionSite = None
        self.cur_idx = 0
        self.max_idx = len(plans) - 1
        self._current_prepare:SitePlan = None

    def next(self) -> SitePlan | None:
        """
        获取下一个建筑点

        如果当前建筑点已经建造完毕或不存在但是有储备建造点，则弹出并返回下一个建筑点

        Returns:
            SitePlan | None: 下一个建筑点，如果没有则返回None
        """
        if self._current_prepare is None and self.cur_idx <= self.max_idx:
            tmp = self.plans[self.cur_idx]
            self.cur_idx += 1
            _pr = tmp.prepare()
            if st.jnumber(_pr):
                return self.next()
            self._current_prepare = tmp

        if not st.site(self.site) and self._current_prepare is not None:
            self.site = self._current_prepare.site
            self._current_prepare = None

        return self.site

    def remove(self):
        """
        移除当前建筑点

        如果存在当前建筑点，则移除它，并清空相关状态
        """
        if st.site(self.site):
            self.site.remove()
        self.site = None

        if self._current_prepare is not None:
            self._current_prepare.site.remove()
            self._current_prepare = None

    def __str__(self):
        return f"SitePlaner({self.plans})"
class EngineMonitor(_ListenSystem):
    name = "EngineMonitor"

    def __init__(self):
        self.gameobjects = set(get.gobjects())

    def on_new(self, objs):
        for obj in objs:
            tpe = get.classname(obj)
            self.pushEvent(f"new-{tpe}", {"obj": obj})

    def on_lost(self, objs):
        for obj in objs:
            tpe = get.classname(obj)
            self.pushEvent(f"lost-{tpe}", {"obj": obj})

    def handle(self):
        _now_objs = set(get.gobjects())
        news = _now_objs.difference(self.gameobjects)
        lost = self.gameobjects.difference(_now_objs)
        self.gameobjects = _now_objs
        # handle
        if news: self.on_new(news)
        if lost: self.on_lost(lost)

