from builtin import *
# utils
class ut:
    VIEW_NAME_OPTS = UsrObject()

    class Stage:
        """
        有限状态机

        .current: str 当前状态

        .next(): str 从当前状态转移到下一个状态

        Args:
            sdef: 一个list对象, 每个元素是一个转移定义: <br/>[src: str | tuple, dst: str | tuple, condition: callable -> bool, action: callable]

                * 每个callable形如 callable(stage:Stage, trans-local:object). stage表示所有当前的Stage实例; trans-local表示当前转移内部的局部空间

                * 如果是固定跳转，那么可以只传入action。即: [src: str | tuple, dst: str | tuple, action: callable]

                !!! 注意: sdef不能是空列表

                * 特别的，src可以是 ＊。当某个src没有任何可行跳转时，会尝试使用 ＊ -> 的跳转判断。

                * 特别的，src和dst可以是tuple。这通常表示二者的condition和action是一样的。

                例如: [('a', 'b'), 'c', ...]将被解析为 ['a', 'b', ...], ['b', 'c', ...]

            entry: 初始状态. 如果不指定, 则默认为sdef[0]的src

        Examples:
            ```python
            spawn = get.spawn(st.friend)

            # 一个Carrier的状态机
            sc = Stage([
                # 任何状态下，如果我没有能量储备了，就去获取Energy
                ['*', 'fetch', lambda s, t: get.energy(s.unit) == 0, lambda s, t: put.fetch(s.unit, get.closest(s.unit, get.boxes(st.energetic)))],
                # 任何状态下，如果我身上有能量，就把能量带回家
                ['*', 'deposit', lambda s, t: get.energy(s.unit) > 0, lambda s, t: put.deposit(s.unit, spawn)],
            ])


            def loop():
                if not sc.unit:
                    res = put.create(spawn, [CARRY, MOVE])
                    sc.unit = res if st.creep(res) else None
                else:
                    print('carrier:', sc.next())
            ```
        """

        def __init__(self, sdef: list, entry: str = None):
            self._raw = sdef
            def_len = len(sdef)
            if def_len == 0:
                raise ValueError("Empty transition definition")

            # 获取状态集 set[str]
            # 获取状态转移字典 dict[src, list[tuple(dst, condition, action)]]  # 若无action, 则为None
            self._trans, self._any_trans, tmp = {}, [], []

            for i in range(len(sdef)):
                item = sdef[i]

                # 这一段是考虑到transcrypt的语法解析问题，不敢写的太风骚: for src, dst, *rest in smdef. 怕出问题
                len_trans = len(item)
                if not 3 <= len_trans <= 4:
                    raise ValueError(f"Invalid transition definition: {sdef[i]}")
                src, dst, cond = item[0], item[1], item[2]
                action = item[3] if len_trans == 4 else None

                if isinstance(src, str):
                    src = [src]

                if isinstance(dst, str):
                    dst = [dst]

                # 形式上写在一起，因此共用一个DataArea(本质就是一个空object)
                da = UsrObject()

                for _src in src:
                    for _dst in dst:
                        if _src != '*':
                            # 写入到常规跳转表
                            stage_node = self._trans.py_get(_src, None)
                            if stage_node:
                                stage_node.append((_dst, cond, action, da))
                            else:
                                self._trans[_src] = [(_dst, cond, action, da)]

                            tmp.extend([_src, _dst])
                        else:
                            self._any_trans.append((_dst, cond, action, da))

            # 再把any_trans添加到所有跳转集的末尾
            for v in self._trans.values():
                v.extend(self._any_trans)

            self._states = set(tmp)
            self._entry = entry if entry else sdef[0][0]
            self.current = self._entry  # 当前状态 str

            # // 更新函数集
            self._su = []  # (attr_name, value_func)
            self._cu = {}  # current: [(attr_name, value_func)]
            self._tu = {}  # (src, dst): [(attr_name, value_func)]

        def asu(self, attr_name: str, value_func: callable):
            """
            Add Stage Update<br/>
            添加属性更新函数<br/>
            用于在状态机的next时自动更新属性<br/>
            * 暂时无法移除，谨慎添加

            Args:
                attr_name: str 属性名
                value_func: callable(s) -> any 更新函数，s为当前Stage实例

            Returns:

            Examples:
                ```python
                carrier = get.friend(lambda obj: obj.name == 'carrier')
                sc = Stage([
                    ['home', 'boxing', lambda s, t: s.energy == 0, lambda s, t: put.fetch(carrier, t.box)],
                    ['boxing', 'home', lambda s, t: s.energy > 0, lambda s, t: put.deposit(carrier, s.spawn)],
                ])

                # update on every next
                sc.asu('energy', lambda s, t: get.energy(carrier))
                # update on every next if current is 'boxing'
                sc.acu('boxing', 'spawn', lambda s, t: get.spawn(st.friend))  # 实际建议将spawn放在全局初始化中, 这里只是举例
                # only update when 'home -> boxing'
                sc.atu('home', 'boxing', 'box', lambda s, t: get.closest(carrier, get.boxes(st.energetic)) )
                ```

            """
            self._su.append((attr_name, value_func))

        def acu(self, current: str | tuple, attr_name: str, value_func: callable):
            """
            Add Current Update<br/>
            添加特定状态下的属性更新函数<br/>
            用于在状态机特定current下的next时自动更新属性<br/>
            * 暂时无法移除，谨慎添加

            Args:
                current: str | tuple 当前状态, 如果是tuple则表示多个状态均添加
                attr_name: str 属性名
                value_func: callable(s) -> any 更新函数，s为当前Stage实例

            Returns:

            Examples:
                ```python
                carrier = get.friend(lambda obj: obj.name == 'carrier')
                sc = Stage([
                    ['home', 'boxing', lambda s, t: s.energy == 0, lambda s, t: put.fetch(carrier, t.box)],
                    ['boxing', 'home', lambda s, t: s.energy > 0, lambda s, t: put.deposit(carrier, s.spawn)],
                ])

                # update on every next
                sc.asu('energy', lambda s, t: get.energy(carrier))
                # update on every next if current is 'boxing'
                sc.acu('boxing', 'spawn', lambda s, t: get.spawn(st.friend))  # 实际建议将spawn放在全局初始化中, 这里只是举例
                # only update when 'home -> boxing'
                sc.atu('home', 'boxing', 'box', lambda s, t: get.closest(carrier, get.boxes(st.energetic)) )
                ```
            """
            if isinstance(current, str):
                current = [current]
            for c in current:
                v = self._cu.py_get(c, None)
                if not v:
                    v = []
                    self._cu[c] = v
                v.append((attr_name, value_func))

        def atu(self, src: str | tuple, dst: str | tuple, attr_name: str, value_func: callable):
            """
            Add Trans Update<br/>
            添加特定转移时的属性更新函数<br/>
            用于在状态机特定 src->dst 时对trans-area中属性的自动更新<br/>
            * 暂时无法移除，谨慎添加

            Args:
                src: str | tuple 当前状态, 如果是tuple则表示多个状态均添加
                dst: str | tuple 目标状态, 如果是tuple则表示多个状态均添加
                attr_name: str 属性名
                value_func: callable(s, t) -> any 更新函数，s为当前Stage实例, t为当前转移的局部空间

            Returns:

            Examples:
                ```python
                carrier = get.friend(lambda obj: obj.name == 'carrier')
                sc = Stage([
                    ['home', 'boxing', lambda s, t: s.energy == 0, lambda s, t: put.fetch(carrier, t.box)],
                    ['boxing', 'home', lambda s, t: s.energy > 0, lambda s, t: put.deposit(carrier, s.spawn)],
                ])

                # update on every next
                sc.asu('energy', lambda s, t: get.energy(carrier))
                # update on every next if current is 'boxing'
                sc.acu('boxing', 'spawn', lambda s, t: get.spawn(st.friend))  # 实际建议将spawn放在全局初始化中, 这里只是举例
                # only update when 'home -> boxing'
                sc.atu('home', 'boxing', 'box', lambda s, t: get.closest(carrier, get.boxes(st.energetic)) )
                ```
            """
            if isinstance(src, str):
                src = [src]
            if isinstance(dst, str):
                dst = [dst]
            for s in src:
                for d in dst:
                    v = self._tu.py_get((s, d), None)
                    if not v:
                        v = []
                        self._tu[(s, d)] = v
                    v.append((attr_name, value_func))

        def next(self, log: bool = True) -> str:
            """
            从当前状态转移到下一个状态

            Args:
                log: 是否返回log信息

            Returns:
                str: log=False: 下一个状态. 如果没有进行转移, 则返回当前状态 | log=True: 返回一句跳转信息: 'src -> dst'
            """

            # 更新su
            for a_name, vfn in self._su:
                setattr(self, a_name, vfn(self, None))

            # 更新cu
            v = self._cu.py_get(self.current, None)
            if v:
                for a_name, vfn in v:
                    setattr(self, a_name, vfn(self, None))

            tset = self._trans.py_get(self.current)
            tset = tset if tset else self._any_trans
            for dst, cond, action, trans_local in tset:
                v = self._tu.py_get((self.current, dst), None)
                if v:
                    for a_name, vfn in v:
                        setattr(trans_local, a_name, vfn(self, trans_local))

                try:
                    cond_res = bool(cond(self, trans_local))
                except Exception as e:
                    raise ValueError(f"[Stage] {self.current} -> {dst}: Usr Condition error: \n{e}")

                if not action:
                    cond_res = True

                if cond_res:
                    if action:
                        try:
                            action(self, trans_local)
                        except Exception as e:
                            raise ValueError(f"[Stage] {self.current} -> {dst}: Usr Action error: \n{e}")
                    info = self.current + " -> " + dst
                    self.current = dst
                    if log:
                        return info
                    return dst
            if log:
                return self.current + " -> " + self.current
            return self.current

        def copy(self):
            """
            '复制'一个Stage实例

            * 只会复制当前sdef和各种update函数，不会复制当前状态和局部空间数据

            Returns:
                Stage

            """
            s = ut.Stage(self._raw)
            s._su = self._su.copy()
            s._cu = self._cu.copy()
            s._tu = self._tu.copy()
            return s

        def clear_(self):
            self.current = self._entry

    class View:
        """
        Visual的包装类

        Args:
            layer: int  视图层级
            persistent: bool  是否持久化视图

        Examples:
            ```python
            view = ut.View(1, True)
            view.clear().text('Hello', st.point(0, 0), 0.5, '#FF0000')  # > ignore
            ```
        """

        def __init__(self, layer: int, persistent: bool = True):
            self.v = object
            # > insert self.v = new Visual(layer, persistent);

        def clear(self) -> object:
            """
            清空当前视图

            * 如果编译器提出警告，可以使用 `# > ignore` 来忽略

            Returns:
                View
            """
            self.v.clear()  # > ignore
            return self

        def text(self, text: str, pos: st.point, font: str | int | float, color: str, options: UsrObject = None) -> object:
            """
            绘制文字

            Args:
                text: str  文字内容
                pos: st.point  文字位置
                font: str | int | float  字体大小 可以是 0.5 '12px' 这样的 (优先级高于options.font)
                color: str  颜色  比如 '#FF0000' (优先级高于options.color)
                options: 详见ScreepsArena文档

            Returns:
                View
            """
            options = options or UsrObject()
            options.font = font
            options.color = color
            self.v.text(text, pos, options)
            return self

        def circle(self, pos: st.point, radius: float, fill: str = None, options: UsrObject = None) -> object:
            """
            绘制圆形

            Args:
                pos: st.point  圆心坐标
                radius: float  半径 (单位为方块) (优先级高于options.radius)
                fill: str  填充颜色  比如 '#FF0000' (优先级高于options.fill)
                options: 详见ScreepsArena文档

            Returns:
                View
            """
            options = options or UsrObject()
            options.radius = radius
            options.fill = fill
            self.v.circle(pos, options)
            return self

        def line(self, pos1: st.point, pos2: st.point, width: float = 0.1, color: str = None, options: UsrObject = None) -> object:
            """
            绘制线段

            Args:
                pos1: st.point  起点坐标
                pos2: st.point  终点坐标
                width: float  线宽 (单位为方块) (优先级高于options.width)
                color: str  颜色 比如 '#FF0000' (优先级高于options.color)
                options: 详见ScreepsArena文档

            Returns:
                View
            """
            options = options or UsrObject()
            options.width = width
            options.color = color
            self.v.line(pos1, pos2, options)
            return self

        def rect(self, pos: st.point, width: float, height: float, fill: str = None, options: UsrObject = None) -> object:
            """
            绘制矩形

            Args:
                pos: st.point  左上角坐标
                width: float  宽度
                height: float  高度
                fill: str  填充颜色 比如 '#FF0000' (优先级高于options.fill)
                options: 详见ScreepsArena文档

            Returns:
                View
            """
            options = options or UsrObject()
            options.fill = fill
            self.v.rect(pos, width, height, options)
            return self

        def poly(self, points: list[st.point], fill: str = None, options: UsrObject = None) -> object:
            """
            绘制多边形

            Args:
                points: list[st.point]  多边形的顶点坐标
                fill: str  填充颜色 比如 '#FF0000' (优先级高于options.fill)
                options: 详见ScreepsArena文档

            Returns:
                View
            """
            options = options or UsrObject()
            options.fill = fill

            # > insert var array = [];
            # > insert for (var i = 0; i < len(points); i++) {
            # > insert     array.push(points[i]);
            # > insert }

            self.v.poly(array, options)
            return self

        def size(self) -> int:
            """
            返回视图的大小<br/>
            * 详见ScreepsArena文档

            Returns:
                int
            """
            return self.v.size()

        def header(self, creep: st.creep) -> object:
            """
            为一个Creep绘制标题

            Args:
                creep: st.creep  一个Creep对象

            Returns:
                View
            """
            if creep:
                txt = creep.name or ""
                hp_percent = get.health(creep, True)  # int:0 - 100
                _hex = Math.floor(hp_percent * 255 / 100).toString(16).padStart(2, '0')
                color = f'#AE{_hex}80'
                txt += f"|{hp_percent}"
                self.text(txt, get.offset(creep, 0, -0.6), 0.4, color, ut.VIEW_NAME_OPTS)
            else:
                std.warn("View.header", "No creep object:", creep)
            return self

    class Efr:
        """
        事件驱动框架(精简版)(?青春版)

        Args:
            name: str  框架名称, 用于标识框架。名称重复不会报错，但会影响你debug。 如果不指定，将使用默认名称+id
            step: int  框架的执行步长。默认为None，表示不限制步长. 如果你的峰值事件数量较多，可以考虑限制step
            log: bool  是否开启日志输出。默认为True

        Examples:
            ```python
            efr = ut.Efr('main', None, True)
            efr.login('logger', None, lambda e: print(e.task))  # 注册一个事件工作站，用于输出日志
            efr.push('tmp', 'logger', 'event task data. like:"hello, world!".')  # 推送一个事件, 发起者为'tmp', 目标为'logger'
            ```
        """
        id = 0

        class SolutionMissing(Exception):
            def __init__(self, *args, event=None):
                super().__init__(*args)
                self.event = event

            def __str__(self):
                txt = '\n[SolutionMissing]: ' + "Event refused by all stations.\n"
                if self.event:
                    txt += "\tEvent: " + str(self.event)
                if self.args:
                    txt += "\tDetails:"
                    for info in self.args:
                        txt += "\n\t\t" + str(info)
                return txt

        class Event:
            """
            事件对象

            事件通常有5个状态：
            1.STATE_OFFLINE - 游离态 0
                刚刚实例化的事件对象，尚未添加到事件框架中。
            2.STATE_JUNIOR - 预备态 1
                该事件被EventFramework.push()到了事件队列。但尚未被进一步处理的状态
            3.STATE_URGENT - 激发态 2
                该事件已经被分配到了EventStation中，但尚未被执行的状态
            4.STATE_FINISH - 完成态 3
                该事件至少被成功执行了一次
                此时的Event可以通过Event.result获取部分结果
            5.STATE_RETIRED - 末态 4
                该事件被所有工作站成功执行, 并且已经脱离了EventFramework
                此时的Event可以通过Event.result获取所有结果
            *6.STATE_EXCEPT - 异常态 -1
                该事件在任意环节中出错，通过Event.trace获取更多错误信息

            The event generally has four states:
            1.STATE_OFFLINE - free state
                The event object just instantiated has not been added to the EventFramework.
            2.STATE_JUNIOR - ready
                The event was by EventFramework.push() to the event queue. A state that has not yet been further processed
            3.STATE_URGENT - excited state
                The state in which the event has been assigned to an EventStation but has not yet been executed
            4.STATE_FINISH - finish atleast once
                The event was executed successfully atleast once and still in EventFramework
                At this time, the event can be through Event.result get the part of the results
            5.STATE_RETIRED - final state
                The event was executed successfully in all task and has broken away from the EventFramework
                At this time, the event can be through Event.result get the all the results
            *6.STATE_EXCEPT - exception
                The event has an error in any step, which is passed through Event.trace for more error information

            Args:
                source: str | object  事件发起者名称。
                dest: str | object  事件目标名称。通常情况下建议使用字符串，使用object时，请确认filter函数的正确性
                task: object  事件任务数据。通常用于传递任务内容数据
                tags: list  事件标签 通常用于标记事件的类型，便于您设计filter函数
                times: int  事件最大执行次数，默认为65535，必须大于0。如果<=1，则表示最多只允许执行一次

            """
            # static const
            STATE_OFFLINE = 0
            STATE_JUNIOR = 1
            STATE_URGENT = 2
            STATE_FINISH = 3
            STATE_RETIRED = 4
            STATE_EXCEPT = -1

            EVENT_ONCE = 1
            EVENT_ALL = 65535

            def __init__(self, source: str | object = None, dest: str | object = None, task: object = None, tags: list = None, times: int = 65535):
                self.efr = None

                self.trace = []
                self.task = task
                self.result = {}
                self.source = source
                self.dest = dest
                self.tags = tags if tags is not None else ()
                self.state = self.STATE_OFFLINE
                self.times = times
                self.links: int = 0

            def offline(self) -> bool:
                """
                检查事件是否处于游离状态</br>
                Check whether the event is offline

                Returns:
                    如果是游离状态返回True, 否则返回False
                """
                return self.state == self.STATE_OFFLINE

            def junior(self) -> bool:
                """
                检查事件是否已经被添加到事件队列中或已经处于urgent状态</br>
                Check whether the event has been added to eventQueue

                Returns:
                    如果是junior或urgent状态返回True, 否则返回False
                """
                return self.state == self.STATE_JUNIOR or self.state == self.STATE_URGENT

            def urgent(self) -> bool:
                """
                检查事件是否已经被添加到station中</br>
                Check whether the event has been added to station

                Returns:
                    如果是urgent状态返回True, 否则返回False
                """
                return self.state == self.STATE_URGENT

            def finish(self) -> bool:
                """
                检查事件是否已经结束或完全结束。只有结束的事件才可以获取result</br>
                Check whether the event has finished or retired. Only the finished event can get the result.

                Returns:
                    如果是finish或retired状态返回True, 否则返回False
                """
                return self.state == self.STATE_FINISH or self.state == self.STATE_RETIRED

            def retired(self) -> bool:
                """
                检查事件是否已经完全结束。</br>
                Check whether the event has retired.

                Returns:
                    如果是retired状态返回True, 否则返回False
                """
                return self.state == self.STATE_RETIRED

            def excepted(self) -> bool:
                """
                检查事件是否出现错误</br>
                Check whether the event has except.

                Returns:
                    如果是except状态返回True, 否则返回False
                """
                return self.state == self.STATE_EXCEPT

            def excepts(self) -> list[Exception]:
                """
                获取event的错误信息. 尚未出错或失败返回[]</br>
                Get the exception-information of event. No error or failure returned []

                Returns:
                    list: 错误信息列表
                """
                return self.trace

            def set_result(self, station_key: object, result: object, syscall=False) -> None:
                """
                set the result of event</br>
                (API for outer caller? maybe)

                Args:
                    station_key: object
                    result: object 事件结果(通常是工作站的respond返回值)
                    syscall: bool 是否来自EventFramework的调用。如果不是，它将自动设置其状态。(用户不应传入此参数)

                Returns:
                    None
                """
                self.result[station_key] = result
                if not syscall:
                    self.links -= 1
                    if self.links <= 0:
                        self.state = self.STATE_FINISH
                    else:
                        self.state = self.STATE_RETIRED

            def get_result(self, station_key: str = None) -> object:
                """
                get the result of event from station. return Get or None

                Args:
                    station_key: str 事件工作站的key 如果为None则返回第一个搜索到的结果(通常用于唯一工作站的情况)

                Returns:
                    object: 事件结果
                """
                if station_key is not None:
                    return self.result.py_get(station_key)
                else:
                    get = list(self.result.values())
                    return get[0] if get else None

            def __str__(self):
                s_tsk = str(self.task)
                s_tsk = s_tsk if len(s_tsk) < 40 else (s_tsk[:40] + '...')
                txt = "Event[source:{}, dest:{}, task:{}, tags:{}]".format(self.source, self.dest, s_tsk, self.tags)
                return txt

        class EventQueue:
            """
            事件队列，事件框架的事件入口

            EventQueue, event entry of EventFramework

            * 一般情况下，用户不需要直接操作EventQueue，而是通过EventFramework的push方法来推送事件

            Args:
                None
            """

            def __init__(self):
                self.efr = None
                self.equeue = []

            def push(self, event: object) -> object:
                """
                驱动事件。成功返回该event, 失败返回None

                Args:
                    event: Event 事件实例

                Returns:
                    Event: 事件实例
                """
                self.equeue.append(event)
                event.efr = self.efr
                if event.state == ut.Efr.Event.STATE_OFFLINE:
                    event.state = ut.Efr.Event.STATE_JUNIOR
                return event

            def release(self, num: int | None = None) -> list:
                """
                获取事件队列中的事件, 该函数不推荐用户使用。用户调用会导致EventFramework工作异常</br>
                Get events in the event queue. This function is not recommended for users. User calls will cause the EventFramework to work abnormally

                Args:
                    num: int 一次获取的事件数量。返回的事件数量不会超过此值(如果为None则表示无限制)

                Returns:
                    list: 事件列表
                """
                if num is None or num >= len(self.equeue):
                    _ = self.equeue
                    self.equeue = []
                    return _

                _ = self.equeue[:num]
                self.equeue = self.equeue[num:]
                return _

            def __len__(self):
                return len(self.equeue)

        class EventStation:
            """
            事件工作站

            # 重要说明:
                小心死锁的情况(A调用(push)到B，但是B必须在A执行完后再能返回)

            Args:
                key: str 工作站的key
                filter: callable 事件筛选函数 如果为None则当event.dest == key时会接收事件
                respond: callable 事件响应函数 如果为None则不响应
                step: int 事件处理步长 默认为None，表示一回合处理完所有事件
                lv: int 工作站优先级(默认为0)(越大越优先)
            """

            def __init__(self, key: str, filter: callable = None, respond: callable = None, step: int = None, lv: int = None):
                self.efr = None
                self.key = key

                self.equeue = ut.Efr.EventQueue()

                self._filter = filter
                self._respond = respond

                self.step = step
                self.lv = 0 if lv is None else lv

            # @abstractmethod
            def filter(self, event: object) -> bool:
                """
                如果初始化时未传入filter则需要用户覆盖此方法</br>
                If filter is not passed in during initialization, you need to override this method

                定义工作站对事件的筛选方法. 筛选通过返回True, 失败返回False</br>
                Defines how the workstation filters events. The filter returns true if passed and false if failed

                默认筛选方法是: 判断事件的dest是否和工作站的key匹配</br>
                The default filtering method is to judge whether the dest of the event matches the workstation.key

                Args:
                    event: Event 事件实例

                Returns:
                    bool，True表示接收事件，False表示拒绝事件
                """
                if self._filter:
                    return self._filter(event)
                return self.key == event.dest

            # @abstractmethod
            def respond(self, event: object) -> object:
                """
                如果初始化时未传入filter则需要用户覆盖此方法</br>
                If filter is not passed in during initialization, you need to override this method

                Args:
                    event: Event 事件实例

                Returns:
                    object 事件处理结果
                """
                if self._respond:
                    return self._respond(event)
                return

            def __str__(self):
                txt = f"estation: {self.key}"
                return txt

            def push(self, event: object) -> bool:
                """
                推送事件到工作站队列</br>
                Push event to workstation queue

                Args:
                    event: Event 事件实例

                Returns:
                    bool 事件是否成功推送
                """
                event = self.equeue.push(event)
                if self.efr.log:
                    std.info(f"estation<{self.key}>", 'recv event:', str(event))
                if event:
                    event.result[self.key] = None
                    if event.state == ut.Efr.Event.STATE_JUNIOR:
                        event.state = ut.Efr.Event.STATE_URGENT
                    event.links += 1
                    return True
                return False

            def update(self):
                events = self.equeue.release(self.step)
                for i, event in enumerate(events):
                    if event.state != ut.Efr.Event.STATE_EXCEPT:
                        try:
                            ret = self.respond(event)
                        except Exception as err:
                            event.state = ut.Efr.Event.STATE_EXCEPT
                            err.where = f"respond of estation: {self.key}"
                            event.trace.append(err)
                            if self.efr.log:
                                std.log(f"esta<{self.key}>", str(err))
                            continue

                        if self.efr.log:
                            std.log(f"esta<{self.key}>", 'handle:', str(event))

                        # running
                        if ret is None:
                            # skip it:
                            if self.efr.log:
                                std.log(f"esta<{self.key}>", 'keep running:', str(event))
                            self.equeue.equeue.append(event)
                            continue
                        elif ret is undefined:
                            if self.efr.log:
                                std.log(f"esta<{self.key}>", 'break on running:', str(event))
                            self.equeue.equeue.extend(events[i:])
                            break

                        if event.state != ut.Efr.Event.STATE_EXCEPT:
                            event.set_result(self.key, ret, syscall=True)
                            event.links -= 1
                            if event.links <= 0:
                                event.state = ut.Efr.Event.STATE_RETIRED
                            else:
                                event.state = ut.Efr.Event.STATE_FINISH

        class EventAlloter:
            """
            事件分配器</br>
            事件分配器会在update函数被调用时工作，取出事件队列中的事件，然后轮询所有已注册的工作站(根据事件的性质可能会提前结束). 最后将事件分配到对应的工作站的队列中

            Event Allocator</br>
            The event allocator will work when the 'update' function is called, take out the events in the event queue, and then poll all registered workstations (depending on the nature of the event, it may end in advance) Finally, the event is allocated to the queue of the corresponding workstation

            Args:
                equeue: object 事件队列
                step: int 事件分配器的步长。表示每次update时，分配器最多分配多少个事件。默认为None，表示不限制步长
            """

            def __init__(self, equeue: object, step: int = None):
                self.efr = None

                self.equeue = equeue
                self.step = step

                self.stations = []

            def login(self, station: object) -> bool:
                """
                注册事件工作站。当update时，alloter会用调用station.filter(事件)->bool来判断station是否响应此事件。如果station响应了事件，那么alloter会把事件传入该station的队列.</br>
                成功返回True，失败返回False

                Register the event workstation. When updating, the alloter will call station Filter - > bool to determine whether the station responds to this event. If the station responds to an event, the alloter will pass the event into the queue of the station</br>
                Returns true for success and false for failure

                Args:
                    station: object 工作站实例

                Returns:
                    bool 事件是否成功注册
                """
                for my_station in self.stations:
                    if my_station.key == station.key:
                        return False

                station.efr = self.efr

                for i in range(len(self.stations)):
                    if station.lv > self.stations[i].lv:
                        self.stations.insert(i, station)
                        return True
                self.stations.append(station)
                return True

            def logoff(self, station: object) -> bool:
                """
                注销工作站。成功或不存在station返回True，失败返回False</br>
                Log off the eventstation. Successful or non-existent stations return true, while failure returns false

                Args:
                    station: object 工作站实例

                Returns:
                    bool 事件是否成功注销
                """
                try:
                    station.etr = None
                    self.stations.remove(station)
                except:
                    return False

                return True

            def update(self):
                """
                由外部单元执行更新</br>
                Updates are performed by external units

                """
                events = self.equeue.release(self.step)

                for event in events:
                    flag = False
                    for station in self.stations:
                        # try filter.
                        try:
                            get = station.filter(event)
                        except Exception as err:
                            event.state = ut.Efr.Event.STATE_EXCEPT
                            err.where = f"filter of estation: {station.key}"
                            event.trace.append(err)
                            if self.efr.log:
                                std.warn(f"alloter: {self.efr.name}", str(err))
                            continue

                        # add to station
                        if get:
                            get = station.push(event)
                            if get:
                                flag = True

                            event.times -= 1
                            if event.times <= 0:
                                break

                    # set exception
                    if not flag:
                        event.state = ut.Efr.Event.STATE_EXCEPT
                        # record except
                        err = ut.Efr.SolutionMissing(event)
                        err.where = f"alloter: {self.efr.name}"
                        event.trace.append(err)
                        if self.efr.log:
                            std.warn(f"alloter: {self.efr.name}", str(err))

        def __init__(self, name: str = None, step: int = None, log=True):
            self.name = name if name else f"EventFramework-{ut.Efr.id}"
            ut.Efr.id += 1

            self.log = log

            # 外部组件
            self.equeue = ut.Efr.EventQueue()
            self.ealloter = ut.Efr.EventAlloter(self.equeue, step)

            # 绑定
            self.equeue.efr = self
            self.ealloter.efr = self

        @property
        def stations(self) -> list:
            """
            获取所有已经注册过的station

            Get all registered stations

            Returns:
                list: 工作站列表
            """
            return self.ealloter.stations

        def login(self, key: str | EventStation, filter=None, respond=None, step: int = None, lv: int = None) -> bool:
            """
            注册事件工作站。当update时，alloter会用调用station.filter(事件)->bool来判断station是否响应此事件。如果station响应了事件，那么alloter会把事件传入该station的队列.</br>
            成功返回True，失败返回False

            Register the event workstation. When updating, the alloter will call station Filter - > bool to determine whether the station responds to this event. If the station responds to an event, the alloter will pass the event into the queue of the station</br>
            Returns true for success and false for failure

            Args:
                key: str | EventStation 工作站key 或 工作站实例
                filter: callable 事件筛选函数 如果为None则当event.dest == key时会接收事件
                respond: callable 事件响应函数 如果为None则不响应
                step: int 事件处理步长 默认为None，表示一回合处理完所有事件
                lv: int 工作站优先级(默认为0)(越大越优先)

            Returns:
                bool 事件是否成功注册
            """
            if isinstance(key, str):
                station = ut.Efr.EventStation(key, filter, respond, self.ealloter.step if step is None else step, lv)
            else:
                station = key

            # 绑定与内绑定
            station.efr = self
            station.equeue.efr = self

            return self.ealloter.login(station)

        def logoff(self, key: str) -> bool:
            """
            注销工作站。成功或不存在station返回True，失败返回False</br>
            Log off the eventstation. Successful or non-existent stations return true, while failure returns false

            Args:
                key: str 工作站key

            Returns:
                bool 事件是否成功注销
            """
            for station in self.estations:
                if station.key == key:
                    return self.ealloter.logoff(station)
            return False

        def push(self, src: str, dst: str, task: object = None, tags: list = None, times: int = 65535) -> Event:
            """
            驱动事件。成功返回该event, 失败返回None

            Args:
                src: str 事件发起者
                dst: str 事件目标
                task: object 事件任务
                tags: list 事件标签
                times: int 事件最大执行次数，默认为65535，必须大于0。如果<=1，则表示最多只允许执行一次

            Returns:
                Event: 事件实例
            """
            event = ut.Efr.Event(src, dst, task, tags, times)
            if self.log:
                std.info(f"efr.push: ", str(event))
            return self.equeue.push(event)

        def update(self, update_stations=True) -> None:
            """
            事件框架更新

            efr update

            Args:
                update_stations: bool 是否更新工作站

            Returns:
                None
            """
            self.ealloter.update()
            if update_stations:
                for station in self.ealloter.stations:
                    station.update()

        def __str__(self):
            return 'EventFramework'

    class Mp:
        """
        The map class, used to store the terrain information of the map and pooling maps

        * should only initialize once in the main loop at the first tick.

        Args:
            width: int, the width of the map, default 100
            height: int, the height of the map, default 100
        """

        def __init__(self, width=100, height=100):
            _map = []
            for ih in range(height):
                new_line = []
                for iw in range(width):
                    ter = get.terrain(iw, ih)
                    if ter == TERRAIN_PLAIN:
                        new_line.append(2)
                    elif ter == TERRAIN_SWAMP:
                        new_line.append(10)
                    else:
                        new_line.append(255)
                _map.append(new_line)
            self.map = _map
            self.width = width
            self.height = height

            self.p2 = self.pooling(self.map, 2)
            self.p4 = self.pooling(self.p2, 2)
            self.p5 = self.pooling(self.map, 5)
            self.p8 = self.pooling(self.p4, 2)
            self.p10 = self.pooling(self.p5, 2)

        def get(self, x: int, y: int) -> int:
            """
            Get the terrain value of the map

            Args:
                x: int, the x position of the map
                y: int, the y position of the map

            Returns:
                int, the terrain value of the map
            """
            return self.map[y][x]

        def set(self, x: int, y: int, value: int) -> None:
            """
            Set the terrain value of the map

            Args:
                x: int, the x position of the map
                y: int, the y position of the map
                value: int, the terrain value of the map
            """
            self.map[y][x] = value

        @staticmethod
        def pooling(tar_arr, kernel_size=2) -> list[list[int]]:
            """
            Pooling the arr with kernel_size and stride with 'mean'

            Args:
                tar_arr: list[list[int]]
                kernel_size: int, strides will equal to kernel_size

            Returns:
                池化后的数组
            """
            if kernel_size < 2:
                raise ValueError("Kernel size should be greater than 1")
            if len(tar_arr) < kernel_size or len(tar_arr[0]) < kernel_size:
                raise ValueError("Array size should be greater than kernel size")
            res = []
            k_square = kernel_size ** 2
            _len_h = (len(tar_arr) // kernel_size) * kernel_size
            _len_w = (len(tar_arr[0]) // kernel_size) * kernel_size
            for i in range(0, _len_h, kernel_size):
                new_line = []
                for j in range(0, _len_w, kernel_size):
                    _sum = 0
                    for k in range(kernel_size):
                        for l in range(kernel_size):
                            _sum += tar_arr[i + k][j + l]
                    new_line.append(_sum // k_square)

                res.append(new_line)
            return res

        def get_pooling(self, psize: int) -> (list[list[int]], int):
            """
            Get the pooling array with psize

            Args:
                psize: int, the pooling size

            Returns:
                list[list[int]], int
            """
            if psize == 1:
                return self.map, 1
            elif psize == 2:
                return self.p2, 2
            elif psize == 4:
                return self.p4, 4
            elif psize == 5:
                return self.p5, 5
            elif psize == 8:
                return self.p8, 8
            elif psize == 10:
                return self.p10, 10
            else:
                raise ValueError("psize should be 1, 2, 4, 5, 8 or 10. Got: %d" % psize)

        def find_space2x(self, psize: int, x: int, y: int, rot: int = 0) -> Point:
            """
            Find a 2x psize square space with the base point x, y

            Args:
                psize: int, the size of the pool. (1, 2, 4, 5, 8, 10)对应搜寻的space大小为(2x2, 4x4, 8x8, 10x10, 16x16, 20x20)
                x: int, the x position of the 100x100 space
                y: int, the y position of the 100x100 space
                rot: 0/1, 表示螺旋的方向, 0代表左逆时针, 1代表右逆时针

            Returns:
                the lt position of the 2x2 space
            """

            def calc_mean2x2(arr, tx, ty):
                """
                Calculate the mean of the 2x2 area

                Args:
                    arr:
                    tx:
                    ty:

                Returns:
                    int
                """
                _sum = 0
                for i in range(2):
                    for j in range(2):
                        _sum += arr[ty + j][tx + i]
                return _sum // 4

            pool, p_size = self.get_pooling(psize)

            width, height = len(pool[0]), len(pool)
            tix, tiy = x // p_size, y // p_size
            step, dir_id = 1, 0
            best_val, best_pos = 255, Point(x, y)
            if rot:
                directs = [(-1, 0), (0, 1), (1, 0), (0, -1)]
            else:
                directs = [(1, 0), (0, -1), (-1, 0), (0, 1)]

            # // 从目标点开始螺旋外扩遍历，对每个点计算2x2平均值，找到<= 10的pos并返回。找不到就返回tix, tiy
            while True:
                for _ in range(2):
                    for _ in range(step):
                        if (0 <= tix < (width - 1)) and (0 <= tiy < (height - 1)):
                            val = calc_mean2x2(pool, tix, tiy)
                            # print('calc', tix, tiy, '->', val)
                            if val <= 10:
                                return Point(tix * p_size, tiy * p_size)
                            elif val < best_val:
                                best_val = val
                                best_pos = Point(tix * p_size, tiy * p_size)
                        elif (tix < 0 or tix >= width - 1) and (tiy < 0 or tiy >= height - 1):
                            return best_pos
                        tix += directs[dir_id][0]
                        tiy += directs[dir_id][1]
                    dir_id = (dir_id + 1) % 4
                step += 1

        def draw(self, view: Visual, psize: int) -> None:
            """
            在视图上绘制特定池化的图像

            Args:
                view: 视图对象
                psize: 池化大小, 1, 2, 4, 5, 8, 10

            Returns:
                None
            """
            op = UsrObject()
            op.opacity = 0.15
            pool, p_size = self.get_pooling(psize)

            # draw each pool, <= 10: green, >10: yellow, >100: red
            for ih in range(self.height // p_size):
                for iw in range(self.width // p_size):
                    value = pool[ih][iw]
                    if value <= 2:
                        color = '#EEEEEE'
                    elif pool[ih][iw] <= 5:
                        color = '#DDFFAA'
                    elif pool[ih][iw] <= 10:
                        color = '#BBFF88'
                    elif pool[ih][iw] <= 50:
                        color = '#BBFF33'
                    elif pool[ih][iw] <= 100:
                        color = '#EEEE33'
                    elif pool[ih][iw] <= 150:
                        color = '#FFAA33'
                    elif pool[ih][iw] <= 200:
                        color = '#FF6666'
                    else:
                        color = '#FF33AA'
                    view.rect(Point(iw * p_size - 0.5, ih * p_size - 0.5), p_size, p_size, color, op)

        def ray(self, start: Point, direct: str = "x+", end: int = 255) -> int:
            """
            射线检测向一个方向发射射线，直到遇到end值或者边界
            返回射线的长度

            Args:
                start: Point 起点, 计算射线时不会考虑起点的值
                direct: str 方向 "x+", "x-", "y+", "y-"
                end: int 终点值
            """
            if direct == "x+":
                for i in range(start.x, self.width):
                    if self.map[start.y][i] >= end:
                        return i - start.x
                return self.width - start.x
            elif direct == "x-":
                for i in range(start.x, -1, -1):
                    if self.map[start.y][i] >= end:
                        return start.x - i
                return start.x + 1
            elif direct == "y+":
                for i in range(start.y, self.height):
                    if self.map[i][start.x] >= end:
                        return i - start.y
                return self.height - start.y
            elif direct == "y-":
                for i in range(start.y, -1, -1):
                    if self.map[i][start.x] >= end:
                        return start.y - i
                return start.y + 1
            return 0

        def sonar(self, p0: Point, p1: Point, direct: str = "x+", end: int = 255) -> list[int]:
            """
            沿着p0到p1向垂直于线路的一个方向发射声呐，返回每个点的测量距离

            * p0到p1应当与direct垂直，比如x+，那么p0.x == p1.x，否则以p0的x为准

            Args:
                p0: Point 起点x或y
                p1: Point 终点x或y
                direct: str 方向 "x+", "x-", "y+", "y-"
                end: int 视作声呐结束的最小值
            """
            res = []
            if direct == "x+":
                x0 = p0.x
                a, b = p0.y, p1.y
                for y in range(a, b + 1):
                    res.append(self.ray(Point(x0, y), "x+", end))
            elif direct == "x-":
                x0 = p0.x
                a, b = p0.y, p1.y
                for y in range(a, b + 1):
                    res.append(self.ray(Point(x0, y), "x-", end))
            elif direct == "y+":
                y0 = p0.y
                a, b = p0.x, p1.x
                for x in range(a, b + 1):
                    res.append(self.ray(Point(x, y0), "y+", end))
            elif direct == "y-":
                y0 = p0.y
                a, b = p0.x, p1.x
                for x in range(a, b + 1):
                    res.append(self.ray(Point(x, y0), "y-", end))
            return res

        def find_narrow(self, p0:Point, p1:Point, direct:str="x+", length:int=-1, end:int=255) -> dict:
            """
            Find the narrowest length from (p0, p1) towards direct

            Args:
                p0: Point
                p1: Point
                direct: str
                length: int
                end: int

            Returns:
                dict
                    .len: int
                    .list: list[Point]
            """
            min_ = 65535
            min_s_ = []
            min_i_ = -1
            if direct == "x+":
                for x in range(p0.x, self.width):
                    tmp = []
                    if length == 0: break
                    length -= 1
                    for y in range(p0.y, p1.y + 1):
                        if self.map[y][x] < end:
                            tmp.append((x, y))
                    if len(tmp) and len(tmp) < min_:
                        min_ = len(tmp)
                        min_s_ = tmp
                        min_i_ = x
            elif direct == "x-":
                for x in range(p0.x, -1, -1):
                    tmp = []
                    if length == 0: break
                    length -= 1
                    for y in range(p0.y, p1.y + 1):
                        if self.map[y][x] < end:
                            tmp.append((x, y))
                    if len(tmp) and len(tmp) < min_:
                        min_ = len(tmp)
                        min_s_ = tmp
                        min_i_ = x
            elif direct == "y+":
                for y in range(p0.y, self.height):
                    tmp = []
                    if length == 0: break
                    length -= 1
                    for x in range(p0.x, p1.x + 1):
                        if self.map[y][x] < end:
                            tmp.append((x, y))
                    if len(tmp) and len(tmp) < min_:
                        min_ = len(tmp)
                        min_s_ = tmp
                        min_i_ = y
            elif direct == "y-":
                for y in range(p0.y, -1, -1):
                    tmp = []
                    if length == 0: break
                    length -= 1
                    for x in range(p0.x, p1.x + 1):
                        if self.map[y][x] < end:
                            tmp.append((x, y))
                    if len(tmp) and len(tmp) < min_:
                        min_ = len(tmp)
                        min_s_ = tmp
                        min_i_ = y

            return {
                'len': min_,        # narrow的长度
                'list': min_s_,     # narrow的点集
                'pos': min_i_       # narrow的位置
            }







    class CInfo:
        """
        用于存储Creep的Strategic Info</br>
        包括:</br>
        - nums: dict prt_name: num</br>
        - score: int  # 最大评分</br>

        Args:
            creep: Creep
        """

        def __init__(self, creep):
            self.nums = {
                MOVE: 0,
                WORK: 0,
                CARRY: 0,
                ATTACK: 0,
                RANGED_ATTACK: 0,
                HEAL: 0,
                TOUGH: 0,
            }
            prts = get.parts(creep)
            # jprint(prts)
            for prt in prts:
                self.nums[prt] += 1
            self.score = get.score(creep)

    class StV:
        """
        战术视图

        Args:
            draw: int  # 绘制的强度:</br>
                0: 不绘制</br>
                1: 绘制友方Creep信息使能</br>
                2: 绘制敌方Creep信息使能</br>
        """

        SIZE = 10  # 10x10的小块

        @staticmethod
        def _new_arr2d(val: int | list | None = None):
            # 生成一个10x10的空数组
            _empty_list = isinstance(val, list) and len(val) == 0
            new_arr = []
            for _ in range(ut.StV.SIZE):
                if val is not None:
                    row = [val if not _empty_list else [] for _ in range(ut.StV.SIZE)]
                else:
                    row = []
                new_arr.append(row)
            return new_arr

        @staticmethod
        def _new_view2d(layer: int):
            new_arr = []
            for i in range(ut.StV.SIZE):
                new_row = []
                for j in range(ut.StV.SIZE):
                    new_row.append(ut.View(layer, True))
                new_arr.append(new_row)
            return new_arr

        @staticmethod
        def _opacity_fn(strength: int) -> float:
            # 将0-inf的strength转换为0.05-0.5的透明度
            return (1 + 9 * strength / (strength + 200)) * .05

        @staticmethod
        def _get_score(cs_dict: dict, creep, _del=False):
            cst_info = cs_dict.pop(creep.id, None) if _del else cs_dict.py_get(creep.id, None)
            if cst_info is None:
                return get.score(creep)
            return cst_info.score

        @staticmethod
        def ipos(x, y, bia_x: float | int = 0, bia_y: float | int = 0) -> Point:
            return Point(x // 10 + bia_x, y // 10 + bia_y)

        @staticmethod
        def xpos(ix, iy, bia_x: float | int = 0, bia_y: float | int = 0) -> Point:
            return Point(ix * 10 + bia_x, iy * 10 + bia_y)

        def __init__(self, mp, draw: int = 1 | 2, log=False):
            # logging
            self.log = log

            # base data
            self.p5: list[list[int]] = mp.p5  # 20x20的池化地图
            self.p10: list[list[int]] = mp.p10  # 10x10的池化地图
            self.cs = {}  # 统计所有creep的基本数据
            self.ec = {
                'b': self._new_arr2d([]),  # Battle Unit
                'w': self._new_arr2d([]),  # Work Unit
                'o': self._new_arr2d([]),  # Other Unit
                's': self._new_arr2d(0),  # Battle Units的score总和
            }
            self.fc = {
                'b': self._new_arr2d([]),  # Battle Unit
                'w': self._new_arr2d([]),  # Work Unit
                'o': self._new_arr2d([]),  # Other Unit
                's': self._new_arr2d(0),  # Battle Units的score总和
            }

            self.views = self._new_view2d(999)  # 用于绘制的视图
            self.e_spawns = []  # spawn obj, last_spawning_creep
            self.f_spawns = []  # spawn obj, last_spawning_creep

            # find spawns
            for spawn in get.spawns():
                if st.friend(spawn):
                    self.f_spawns.append([spawn, None])
                else:
                    self.e_spawns.append([spawn, None])

            # jprint(self.e_spawns, self.f_spawns)

            # parse draw
            self.draw = bool(draw)
            self.f_draw = draw & 1
            self.e_draw = draw & 2

        def add(self, creep):
            """
            添加creep信息到stv
            Args:
                creep:

            Returns:

            """
            # Creep Info
            cs = ut.CInfo(creep)
            self.cs[creep.id] = cs  # add creep info

            # Friend or Enemy
            is_friend = st.friend(creep)
            ix, iy = creep.x // 10, creep.y // 10
            sel = self.fc if is_friend else self.ec

            # B W O write.
            key = 'o'
            if st.atkable(creep) or st.healable(creep):
                key = 'b'
            elif st.workable(creep):
                key = 'w'

            # write
            sel[key][iy][ix].append(creep)
            sel['s'][iy][ix] += cs.score
            if self.log:
                std.info('ut.StV.add', f'add {"friend" if is_friend else "enemy"} creep:{creep.id}({cs.score}) into "{key}" at ({ix}, {iy}).')

        def draw_area(self, iw, ih):
            """
            绘制指定的地图区块
            Args:
                iw:
                ih:

            Returns:

            """

            fcs, ecs = self.fc['s'][ih][iw], self.ec['s'][ih][iw]
            pos = self.xpos(iw, ih, -.5, .5)
            op = UsrObject()

            if self.log:
                std.info('ut.StV.draw_area', f'draw area ({iw}, {ih}). fcs:{fcs}, ecs:{ecs}. xpos:({pos.x}, {pos.y}).')

            view = self.views[ih][iw]
            view.clear()  # > ignore
            if self.f_draw and fcs:
                op.opacity = self._opacity_fn(fcs)
                view.rect(pos, 10, 10, 'green', op)

            if self.e_draw and ecs:
                op.opacity = self._opacity_fn(ecs)
                view.rect(pos, 10, 10, 'red', op)

        def _move_creeps_and_record(self, select: dict, key: str, iy, ix, moves_record: list, redraw_record: list) -> bool:
            """
            检查creep是否移动到其他区域
            Args:
                select: dict  己方或敌方的creep信息字典
                key: str  'b' 'w' 'o' # Battle, Work, Other
                iy: int current iy
                ix: int current ix
                moves_record: list  # 移动记录
                redraw_record: list  # 需要重新绘制的区域

            Returns:
                如果有creep移动到其他区域，返回True，否则返回False
            """
            scores, area = select['s'], select[key][iy][ix]
            _removes = []
            for i, c in enumerate(area):
                new_ix, new_iy = c.x // 10, c.y // 10
                if c.hits is undefined or c.hits <= 0:  # creep dead(in game 0.18.0 will turn to undefined)
                    _removes.append(i)
                    scores[iy][ix] -= self._get_score(self.cs, c, True)
                    continue
                elif new_ix != ix or new_iy != iy:
                    moves_record.append((c, new_ix, new_iy))
                    _removes.append(i)
                    score = self._get_score(self.cs, c)
                    scores[iy][ix] -= score
                    scores[new_iy][new_ix] += score
                    redraw_record.append((new_ix, new_iy))

            if len(_removes):
                _removes.reverse()
                for i in _removes:
                    area.pop(i)

            return len(_removes) > 0

        def check_new_creep(self, spawn_lcs: list) -> list:
            """
            检查有无新的creep
            Args:
                spawn_lcs:

            Returns:
                返回新的creep列表
            """
            _news = []
            for i in range(len(spawn_lcs)):
                spawn, lc = spawn_lcs[i]
                sc = get.creep()  # > remove
                # > insert var sc = spawn?.spawning?.creep;
                if sc and (not lc or sc.id != lc.id):
                    self.add(sc)
                    _news.append(sc)
                    spawn_lcs[i][1] = sc
            return _news

        def check_creeps_moves(self) -> list:
            """
            检查creep的移动情况
            Args:

            Returns:
                redraw_area_ids: set  # 需要重新绘制的区域
            """
            # check creep move to other area

            # creep: (new_x, new_y)
            _fcb_moves, _ecb_moves = [], []
            _fcw_moves, _ecw_moves = [], []
            _fco_moves, _eco_moves = [], []
            redraw_area_ids = []
            for ih in range(self.SIZE):
                for iw in range(self.SIZE):
                    _ = False
                    if len(self.fc['b'][ih][iw]):
                        _ |= self._move_creeps_and_record(self.fc, 'b', ih, iw, _fcb_moves, redraw_area_ids)
                    if len(self.fc['w'][ih][iw]):
                        _ |= self._move_creeps_and_record(self.fc, 'w', ih, iw, _fcw_moves, redraw_area_ids)
                    if len(self.fc['o'][ih][iw]):
                        _ |= self._move_creeps_and_record(self.fc, 'o', ih, iw, _fco_moves, redraw_area_ids)
                    if len(self.ec['b'][ih][iw]):
                        _ |= self._move_creeps_and_record(self.ec, 'b', ih, iw, _ecb_moves, redraw_area_ids)
                    if len(self.ec['w'][ih][iw]):
                        _ |= self._move_creeps_and_record(self.ec, 'w', ih, iw, _ecw_moves, redraw_area_ids)
                    if len(self.ec['o'][ih][iw]):
                        _ |= self._move_creeps_and_record(self.ec, 'o', ih, iw, _eco_moves, redraw_area_ids)
                    if _:  # redraw
                        redraw_area_ids.append((iw, ih))

            # add new move
            for each in _fcb_moves:
                c, new_x, new_y = each
                self.fc['b'][new_y][new_x].append(c)
            for each in _fcw_moves:
                c, new_x, new_y = each
                self.fc['w'][new_y][new_x].append(c)
            for each in _fco_moves:
                c, new_x, new_y = each
                self.fc['o'][new_y][new_x].append(c)
            for each in _ecb_moves:
                c, new_x, new_y = each
                self.ec['b'][new_y][new_x].append(c)
            for each in _ecw_moves:
                c, new_x, new_y = each
                self.ec['w'][new_y][new_x].append(c)
            for each in _eco_moves:
                c, new_x, new_y = each
                self.ec['o'][new_y][new_x].append(c)
            # finish check move
            return list(set(redraw_area_ids))

        def filter(self, filter_fn: list | tuple | set | callable, ret_pos=True):
            """
            过滤整个二位网络，返回选中[(ix, iy), ...]
            Args:
                filter_fn: list|tuple|set|callable
                    params: crs: dict,
                        crs:{
                            'fb': list[creep],
                            'fw': list[creep],
                            'fo': list[creep],
                            'eb': list[creep],
                            'ew': list[creep],
                            'eo': list[creep],
                            'fs': int,
                            'es': int,
                        }
                ret_pos: bool  # 返回类型为[(ix, iy), ...]，否则返回[[Creep, ], ...]

            Returns:

            """
            filter_fn = std.combo(filter_fn)
            res = []
            for iy in range(self.SIZE):
                for ix in range(self.SIZE):
                    crs = {
                        'fb': self.fc['b'][iy][ix],
                        'fw': self.fc['w'][iy][ix],
                        'fo': self.fc['o'][iy][ix],
                        'eb': self.ec['b'][iy][ix],
                        'ew': self.ec['w'][iy][ix],
                        'eo': self.ec['o'][iy][ix],
                        'fs': self.fc['s'][iy][ix],
                        'es': self.ec['s'][iy][ix],
                    }
                    if filter_fn(crs):
                        if ret_pos:
                            res.append((ix, iy))
                        else:
                            res.append(crs)
            return res

        def find(self, filter_fn: list | tuple | set | callable, ret_pos=True) -> tuple | None:
            """
            查找符合条件的区域
            Args:
                filter_fn: list|tuple|set|callable
                    params: crs: dict,
                        crs:{
                            'fb': list[creep],
                            'fw': list[creep],
                            'fo': list[creep],
                            'eb': list[creep],
                            'ew': list[creep],
                            'eo': list[creep],
                            'fs': int,
                            'es': int,
                        }
                ret_pos: bool  # 返回类型为(ix, iy)，否则返回[Creep, ]

            Returns:
                返回符合条件的区域(ix, iy), 如果找不到则返回None
            """
            filter_fn = std.combo(filter_fn)
            for iy in range(self.SIZE):
                for ix in range(self.SIZE):
                    crs = {
                        'fb': self.fc['b'][iy][ix],
                        'fw': self.fc['w'][iy][ix],
                        'fo': self.fc['o'][iy][ix],
                        'eb': self.ec['b'][iy][ix],
                        'ew': self.ec['w'][iy][ix],
                        'eo': self.ec['o'][iy][ix],
                        'fs': self.fc['s'][iy][ix],
                        'es': self.ec['s'][iy][ix],
                    }
                    if filter_fn(crs):
                        if ret_pos:
                            return ix, iy
                        return crs

            return None

        def crsAt(self, ix: int, iy: int) -> dict:
            return {
                'fb': self.fc['b'][iy][ix],
                'fw': self.fc['w'][iy][ix],
                'fo': self.fc['o'][iy][ix],
                'eb': self.ec['b'][iy][ix],
                'ew': self.ec['w'][iy][ix],
                'eo': self.ec['o'][iy][ix],
                'fs': self.fc['s'][iy][ix],
                'es': self.ec['s'][iy][ix],
            }

        def _update_crs(self, crs, ix, iy):
            if crs is None: return
            crs['fb'] = self.fc['b'][iy][ix]
            crs['fw'] = self.fc['w'][iy][ix]
            crs['fo'] = self.fc['o'][iy][ix]
            crs['eb'] = self.ec['b'][iy][ix]
            crs['ew'] = self.ec['w'][iy][ix]
            crs['eo'] = self.ec['o'][iy][ix]
            crs['fs'] = self.fc['s'][iy][ix]
            crs['es'] = self.ec['s'][iy][ix]

        def rfilter(self, filter_fn: list | tuple | set | callable, start: tuple[int, int] = None, round: int = -1, ret_pos=True) -> list:
            filter_fn = std.combo(filter_fn)
            if start is None:
                start = (self.SIZE // 2, self.SIZE // 2)  # 从中心开始
            x0, y0 = start
            res = []

            if round == -1:
                round = self.SIZE // 2  # 搜索至最大半径 / 2

            crs = {
                'fb': None,
                'fw': None,
                'fo': None,
                'eb': None,
                'ew': None,
                'eo': None,
                'fs': None,
                'es': None,
            }

            for r in range(round + 1):
                # 上
                for k in range(-r, r + 1):
                    ix, iy = x0 + k, y0 + r
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            res.append((ix, iy) if ret_pos else crs)
                # 左
                for k in range(-r, r + 1):
                    ix, iy = x0 - r, y0 + k
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            res.append((ix, iy) if ret_pos else crs)
                # 下
                for k in range(-r, r + 1):
                    ix, iy = x0 + k, y0 - r
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            res.append((ix, iy) if ret_pos else crs)
                # 右
                for k in range(-r, r + 1):
                    ix, iy = x0 + r, y0 + k
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            res.append((ix, iy) if ret_pos else crs)

            return res

        def rfind(self, filter_fn: list | tuple | set | callable, start: tuple[int, int] = None, round: int = -1, ret_pos=True) -> tuple | None:
            filter_fn = std.combo(filter_fn)
            if start is None:
                start = (self.SIZE // 2, self.SIZE // 2)  # 从中心开始
            x0, y0 = start

            if round == -1:
                round = self.SIZE // 2  # 搜索至最大半径 / 2

            crs = {
                'fb': None,
                'fw': None,
                'fo': None,
                'eb': None,
                'ew': None,
                'eo': None,
                'fs': None,
                'es': None,
            }

            for r in range(round + 1):
                # 上
                for k in range(-r, r + 1):
                    ix, iy = x0 + k, y0 + r
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            return (ix, iy) if ret_pos else crs
                # 左
                for k in range(-r, r + 1):
                    ix, iy = x0 - r, y0 + k
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            return (ix, iy) if ret_pos else crs
                # 下
                for k in range(-r, r + 1):
                    ix, iy = x0 + k, y0 - r
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            return (ix, iy) if ret_pos else crs
                # 右
                for k in range(-r, r + 1):
                    ix, iy = x0 + r, y0 + k
                    if 0 <= ix < self.SIZE and 0 <= iy < self.SIZE:
                        self._update_crs(crs, ix, iy)
                        if filter_fn(crs):
                            return (ix, iy) if ret_pos else crs

            return None

        def _check_area_fit(self, ix, iy, type: str) -> bool:
            """
            检查区域是否符合要求
            Args:
                ix:
                iy:
                type:

            Returns:

            """
            if type == 'f':
                return self.fc['s'][iy][ix] > 0
            elif type == 'e':
                return self.ec['s'][iy][ix] > 0
            elif type == 'fb':
                return len(self.fc['b'][iy][ix]) > 0
            elif type == 'fw':
                return len(self.fc['w'][iy][ix]) > 0
            elif type == 'fo':
                return len(self.fc['o'][iy][ix]) > 0
            elif type == 'eb':
                return len(self.ec['b'][iy][ix]) > 0
            elif type == 'ew':
                return len(self.ec['w'][iy][ix]) > 0
            elif type == 'eo':
                return len(self.ec['o'][iy][ix]) > 0
            else:
                return False

        def closest_area(self, ix, iy, type: str, ret_pos=True) -> tuple | None:
            """
            获取最近的区域
            Args:
                ix:
                iy:
                type: f fa fh fo e ea eh eo
                ret_pos: bool # 返回类型为(ix, iy)，否则返回[Creep, ]

            Returns:
                返回最近的区域(ix, iy), 如果找不到则返回None
            """
            if self._check_area_fit(ix, iy, type):
                if ret_pos:
                    return ix, iy
            # 螺旋向外遍历. 对应的list不为空，则返回tuple
            directs = [(0, 1), (1, 0), (0, -1), (-1, 0)]
            step = 1
            while True:
                for i in range(4):
                    for j in range(step):
                        ix += directs[i][0]
                        iy += directs[i][1]
                        if (ix < 0 or ix >= 10) and (iy < 0 or iy >= 10):
                            return None
                        if 0 <= ix < 10 and 0 <= iy < 10 and self._check_area_fit(ix, iy, type):
                            return ix, iy
                step += 1

        def update(self):
            """
            更新地图信息
            """
            # check moves
            redraw = self.check_creeps_moves()

            # for each spawn and check spawning with last tick
            _new_e = self.check_new_creep(self.e_spawns)
            _new_f = self.check_new_creep(self.f_spawns)

            # redraw
            if self.draw:
                for creep in _new_e:
                    self.draw_area(creep.x // self.SIZE, creep.y // self.SIZE)
                for creep in _new_f:
                    self.draw_area(creep.x // self.SIZE, creep.y // self.SIZE)
                for each in redraw:
                    self.draw_area(each[0], each[1])


ut.VIEW_NAME_OPTS.font = '0.5'
ut.VIEW_NAME_OPTS.align = 'center'
ut.VIEW_NAME_OPTS.opacity = 0.7
ut.VIEW_NAME_OPTS.color = '#AEFC80'
